/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include <assert.h>
#include <librnd/core/math_helper.h>
#include <librnd/core/compat_misc.h>

#include "geo.h"


RND_INLINE rnd_bool is_between_d(double val, double lim1, double lim2)
{
	if (lim1 < lim2) return (val >= lim1) && (val <= lim2);
	return (val >= lim2) && (val <= lim1);
}

RND_INLINE rnd_bool is_between_c(rnd_coord_t val, rnd_coord_t lim1, rnd_coord_t lim2)
{
	if (lim1 < lim2) return (val >= lim1) && (val <= lim2);
	return (val >= lim2) && (val <= lim1);
}

RND_INLINE double line_x_for_y(rnd_coord_t y, rnd_coord_t lx1, rnd_coord_t ly1, rnd_coord_t lx2, rnd_coord_t ly2)
{
	double dx = (double)(lx2 - lx1) / (double)(ly2 - ly1);
	return (double)lx1 + (double)(y - ly1) * dx;
}


RND_INLINE double line_y_for_x(rnd_coord_t x, rnd_coord_t lx1, rnd_coord_t ly1, rnd_coord_t lx2, rnd_coord_t ly2)
{
	double dy = (double)(ly2 - ly1) / (double)(lx2 - lx1);
	return ly1 + (double)(x - lx1) * dy;
}


rnd_bool osm_isc_hline_line(rnd_coord_t hx1, rnd_coord_t hy, rnd_coord_t hx2, rnd_coord_t lx1, rnd_coord_t ly1, rnd_coord_t lx2, rnd_coord_t ly2)
{
	double x;

	assert(hx1 < hx2);

	/* quick check: if there is no chance for an intersection */
	if ((hy < (RND_MIN(ly1, ly2))) || (hy > (RND_MAX(ly1, ly2))))
		return rnd_false;

	x = line_x_for_y(hy, lx1, ly1, lx2, ly2);
	return is_between_d(x, hx1, hx2);
}

rnd_bool osm_isc_vline_line(rnd_coord_t vx, rnd_coord_t vy1, rnd_coord_t vy2, rnd_coord_t lx1, rnd_coord_t ly1, rnd_coord_t lx2, rnd_coord_t ly2)
{
	double y;

	assert(vy1 < vy2);

	/* quick check: if there is no chance for an intersection */
	if ((vx < (RND_MIN(lx1, lx2))) || (vx > (RND_MAX(lx1, lx2))))
		return rnd_false;

	y = line_y_for_x(vx, lx1, ly1, lx2, ly2);
	return is_between_d(y, vy1, vy2);
}




/* checks if a line crosses a rectangle or is within the rectangle */
rnd_bool osm_isc_line_rectangle(rnd_coord_t bx1, rnd_coord_t by1, rnd_coord_t bx2, rnd_coord_t by2, rnd_coord_t lx1, rnd_coord_t ly1, rnd_coord_t lx2, rnd_coord_t ly2)
{
	/* if point 1 is inside the rectangle they are intersecting for sure (line in box) */
	if ((bx1 <= lx1) && (bx2 >= lx1) && (by1 <= ly1) && (by2 >= ly1))
		return rnd_true;

	/* box top or bottom horizontal edge */
	if (osm_isc_hline_line(bx1, by1, bx2, lx1, ly1, lx2, ly2))
		return rnd_true;
	if (osm_isc_hline_line(bx1, by2, bx2, lx1, ly1, lx2, ly2))
		return rnd_true;

	/* box left or right vertical edge */
	if (osm_isc_vline_line(bx1, by1, by2, lx1, ly1, lx2, ly2))
		return rnd_true;
	if (osm_isc_vline_line(bx2, by1, by2, lx1, ly1, lx2, ly2))
		return rnd_true;


	return rnd_false;
}

double osm_proj_pt_onto_line(rnd_coord_t *x, rnd_coord_t *y, rnd_coord_t px, rnd_coord_t py, rnd_coord_t lx1, rnd_coord_t ly1, rnd_coord_t lx2, rnd_coord_t ly2)
{
	double ax, ay, dxo, dyo, dx, dy, len, offs;

	/* degenerate case: line is a single point */
	if ((lx1 == lx2) && (ly1 == ly2)) {
		*x = lx1;
		*y = ly1;
		return 0;
	}

	ax = px - lx1; ay = py - ly1;
	dxo = lx2 - lx1; dyo = ly2 - ly1;
	len = sqrt(dxo*dxo + dyo*dyo);

	dx = dxo/len;
	dy = dyo/len;

	offs = (ax * dx + ay * dy) / len;

	*x = rnd_round((double)lx1 + offs * dxo);
	*y = rnd_round((double)ly1 + offs * dyo);

	return offs;
}

