/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include <librnd/config.h>

#include <librnd/core/actions.h>
#include <librnd/core/compat_misc.h>
#include "event.h"

#include "data.h"
#include "plug_io.h"
#include "plug_io_act.h"

static const char *plug_io_cookie = "plug_io_act";

static const char osm_acts_LoadFrom[] = "LoadFrom(Layer|Project,filename)";
static const char osm_acth_LoadFrom[] = "Load project or layer data from a file.";
fgw_error_t osm_act_LoadFrom(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	const char *op, *name;

	RND_ACT_CONVARG(1, FGW_STR, LoadFrom, op = argv[1].val.str);
	RND_ACT_CONVARG(2, FGW_STR, LoadFrom, name = argv[2].val.str);

	if (rnd_strcasecmp(op, "layer") == 0) {
		if (osm_io_load(&osm, name) != 0) {
			rnd_message(RND_MSG_ERROR, "Can not load file '%s'\n", name);
			RND_ACT_IRES(-1);
			return 0;
		}
		rnd_event(&osm.hidlib, OSM_EVENT_LAYERS_CHANGED, NULL);
	}
	else if (rnd_strcasecmp(op, "project") == 0) {
		TODO("the actual project load");
		rnd_message(RND_MSG_ERROR, "LoadFrom(project,...) not yet implemented\n");
	}
	else
		RND_ACT_FAIL(LoadFrom);

	RND_ACT_IRES(0);
	return 0;
}

static const char osm_acts_SaveTo[] = "SaveTo(layer|design, filename[,format])";
static const char osm_acth_SaveTo[] = "Save layer or design to a file.";
fgw_error_t osm_act_SaveTo(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	const char *op, *name, *format = "osm_xml";

	RND_ACT_CONVARG(1, FGW_STR, SaveTo, op = argv[1].val.str);
	RND_ACT_CONVARG(2, FGW_STR, SaveTo, name = argv[2].val.str);
	RND_ACT_MAY_CONVARG(3, FGW_STR, SaveTo, format = argv[3].val.str);

	if (rnd_strcasecmp(op, "design") == 0) {
		if (osm_save_design(osm, name, format) != 0) {
			rnd_message(RND_MSG_ERROR, "Can not save design to file '%s'\n", name);
			RND_ACT_IRES(-1);
			return 0;
		}
	}
	else if (rnd_strcasecmp(op, "layer") == 0) {
		osm_layer_t *ly = osm_layer_get_curr(osm);

		if (ly == NULL) {
			rnd_message(RND_MSG_ERROR, "Layer save: no layer addressed\n");
			RND_ACT_IRES(-1);
			return 0;
		}

TODO("optional arg for selecting layer");
		if (osm_save_layer(osm, ly, name, format) != 0) {
			rnd_message(RND_MSG_ERROR, "Can not save layer to file '%s'\n", name);
			RND_ACT_IRES(-1);
			return 0;
		}
	}
	else {
		rnd_message(RND_MSG_ERROR, "Invalid first argument\n");
		RND_ACT_IRES(-1);
		return 0;
	}

	RND_ACT_IRES(0);
	return 0;
}


static rnd_action_t osm_plug_io_act_list[] = {
	{"LoadFrom", osm_act_LoadFrom, osm_acth_LoadFrom, osm_acts_LoadFrom},
	{"SaveTo", osm_act_SaveTo, osm_acth_SaveTo, osm_acts_SaveTo}
};

void osm_plug_io_act_uninit(void)
{
	rnd_remove_actions_by_cookie(plug_io_cookie);
}

void osm_plug_io_act_init(void)
{
	RND_REGISTER_ACTIONS(osm_plug_io_act_list, plug_io_cookie);
}
