/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#ifndef OSM_PROJECTION_H
#define OSM_PROJECTION_H

#include <math.h>
#include <librnd/core/compat_misc.h>

#define EARTH_R 6378137.0
#define EARTH_C (2.0 * M_PI * EARTH_R)


#define DEG2RAD(a)   ((a) / (180 / M_PI))
#define RAD2DEG(a)   ((a) * (180 / M_PI))

/* Mercator projection assuming Earth is a sphere.
   Since rnd_coord_t is signed 32 bit by default, the origin at 0;0 is both
   the origin of rnd_coords and lon;lat. One unit of rnd_coord_t is one
   centimeter, this way +-180.0 deg longitude coordinates fit.
   For deg<->meter, see: https://wiki.openstreetmap.org/wiki/Mercator#C
*/
RND_INLINE void lldeg2rnd(double lon_deg, double lat_deg, rnd_coord_t *x, rnd_coord_t *y)
{
	double lon = (lon_deg)*M_PI/180.0;
	double lat = (lat_deg)*M_PI/180.0;

	*x = rnd_round(EARTH_R * lon * 100.0);
	*y = rnd_round(log(tan(lat / 2.0 + M_PI/4)) * EARTH_R * 100.0);
}

RND_INLINE void rnd2lldeg(double *lon_deg, double *lat_deg, rnd_coord_t x, rnd_coord_t y)
{
	*lon_deg = RAD2DEG((double)x/100.0/EARTH_R);
	*lat_deg = RAD2DEG(2.0 * atan(exp((double)y/100.0/EARTH_R)) - M_PI/2.0);
}

RND_INLINE double rnd2londeg(rnd_coord_t x)
{
	return RAD2DEG((double)x/100.0/EARTH_R);
}

RND_INLINE double rnd2latdeg(rnd_coord_t y)
{
	return RAD2DEG(2.0 * atan(exp((double)y/100.0/EARTH_R)) - M_PI/2.0);
}

RND_INLINE int osm_nlz2(unsigned x) /* from Hacker's Delight */
{
	unsigned y;
	int n = 32;
	y = x >>16;  if (y != 0) {n = n -16;  x = y;}
	y = x >> 8;  if (y != 0) {n = n - 8;  x = y;}
	y = x >> 4;  if (y != 0) {n = n - 4;  x = y;}
	y = x >> 2;  if (y != 0) {n = n - 2;  x = y;}
	y = x >> 1;  if (y != 0) return n - 2;
	return n - x;
}

RND_INLINE int osm_ilog2(unsigned x)
{
	return 32-osm_nlz2(x);
}

/* Compute an integer zoom level (1..23) from screen coord_per_pix value */
RND_INLINE int osm_zoom_rnd2level(double coord_per_pix)
{
	double meter_per_pixel = coord_per_pix/100.0;
	double tmp = EARTH_C / meter_per_pixel;
	return osm_ilog2((unsigned)tmp)-8;
}

#endif
