/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <librnd/core/config.h>
#include <librnd/core/actions.h>
#include <librnd/hid/hid.h>

#include "conf_core.h"
#include "data.h"
#include "obj_any.h"
#include "search.h"
#include "draw.h"
#include "event.h"

#include "select.h"

void osm_obj_unselect(osm_any_obj_t *o)
{
	o->hdr.selected = 0;
}

void osm_obj_select(osm_any_obj_t *o)
{
	o->hdr.selected = 1;
}

static void sel_chg_notify(osm_layer_t *ly)
{
	rnd_event(&ly->parent->hidlib, OSM_EVENT_SELECTION_CHANGED, NULL);
}


void osm_unselect_all(osm_layer_t *ly)
{
	htIp_entry_t *e;
	int t;

	for(t = 0; t < OSM_OBJ_max; t++)
		for(e = htIp_first(&ly->id2obj[t]); e != NULL; e = htIp_next(&ly->id2obj[t], e))
			osm_obj_unselect(e->value);

	sel_chg_notify(ly);
}

void osm_select_click(osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y)
{
	osm_any_obj_t *obj;
	
	if (rnd_gui->shift_is_pressed(rnd_gui))
		obj = osm_search_obj_xy(ly, x, y); /* for the shift+click unselect */
	else
		obj = osm_search_obj_xy_unsel(ly, x, y);

	if (obj == NULL) { /* clicked away or clicked on selected */
		if (!rnd_gui->shift_is_pressed(rnd_gui))
			osm_unselect_all(ly);
		return;
	}

	if (obj->hdr.selected) {
		if (rnd_gui->shift_is_pressed(rnd_gui))
			osm_obj_unselect(obj);
		else
			osm_obj_select(obj);
	}
	else {
		if (!rnd_gui->shift_is_pressed(rnd_gui))
			osm_unselect_all(ly);
		osm_obj_select(obj);
	}
	sel_chg_notify(ly);
}


void osm_select_box(osm_layer_t *ly, rnd_coord_t x1, rnd_coord_t y1, rnd_coord_t x2, rnd_coord_t y2)
{
	int neg = 0;
	vtp0_t res = {0};
	rnd_coord_t slop = osm_slop;
	osm_rtree_box_t rbox;
	long n;

	osm_draw_inhibit_inc();

	if (!rnd_gui->shift_is_pressed(rnd_gui))
		osm_unselect_all(ly);

	/* figure if selection is negative */
	if (!conf_core.editor.selection.disable_negative) {
		int negx = x2 < x1, negy = y1 < y2; /* y is inverted on screen */

		if (!conf_core.editor.selection.symmetric_negative)
			neg = negy;

		neg |= negx;
	}

	/* make the box canonical */
	if (x1 > x2) rnd_swap(rnd_coord_t, x1, x2);
	if (y1 > y2) rnd_swap(rnd_coord_t, y1, y2);

	rbox.x1 = x1 - slop; rbox.y1 = y1 - slop;
	rbox.x2 = x2 + slop; rbox.y2 = y2 + slop;
	osm_search_append_objs_box(&res, ly, &rbox);

	for(n = 0; n < res.used; n++) {
		osm_any_obj_t *o = res.array[n];

		if (neg) { /* select anything that intersects */
			if (o->hdr.calls->isc_box(o, &rbox))
				osm_obj_select(o);
		}
		else { /* select anything that is fully within */
			if ((o->hdr.bbox.x1 >= rbox.x1) && (o->hdr.bbox.y1 >= rbox.y1) && (o->hdr.bbox.x2 <= rbox.x2) && (o->hdr.bbox.y2 <= rbox.y2))
				osm_obj_select(o);
		}
	}

	sel_chg_notify(ly);

	vtp0_uninit(&res);

	osm_draw_inhibit_dec();
}

static const char osm_acts_Unselect[] = "Unselect(All)";
static const char osm_acth_Unselect[] = "Unselect all objects\n";
fgw_error_t osm_act_Unselect(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);

	osm_unselect_all(ly);

	RND_ACT_IRES(0);
	return 0;
}

static rnd_action_t select_action_list[] = {
	{"Unselect", osm_act_Unselect, osm_acth_Unselect, osm_acts_Unselect}
};


void osm_select_init2(void)
{
	RND_REGISTER_ACTIONS(select_action_list, NULL);
}

