/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <librnd/config.h>

#include <stdlib.h>
#include <string.h>

#include <librnd/core/actions.h>
#include <librnd/hid/hid_dad.h>
#include <librnd/hid/hid_dad_tree.h>

#include "data.h"
#include "obj_any.h"
#include "search.h"
#include "tags.h"

#include "dlg_tags.h"

typedef struct{
	RND_DAD_DECL_NOINIT(dlg)
	int wlist, wobjinfo, wsuggest, wval, wcomment;

	/* object address (or commit) */
	osm_layer_t *ly;
	osm_objtype_t obj_type;
	osm_id_t obj_id;
	osm_tags_t *commit_tags;
} tags_ctx_t;

static void tags_close_cb(void *caller_data, rnd_hid_attr_ev_t ev)
{
	tags_ctx_t *ctx = caller_data;

	if (ctx->commit_tags != NULL) {
		rnd_hid_attribute_t *atxt = &ctx->dlg[ctx->wcomment];
		rnd_hid_text_t *txt = atxt->wdata;
		char *txtstr = txt->hid_get_text(atxt, ctx->dlg_hid_ctx);
		htsp_set(ctx->commit_tags, rnd_strdup("comment"), txtstr);
		txtstr = NULL; /* ownership passed on */
	}

	RND_DAD_FREE(ctx->dlg);
	free(ctx);
}

int tag_cmp(const void *A, const void *B)
{
	htsp_entry_t * const *a = A, * const *b = B;
	int res = strcmp((*a)->key, (*b)->key);
	if (res == 0) return 1;
	return res;
}

static void tags_osm2dlg(tags_ctx_t *ctx)
{
	osm_any_obj_t *obj = NULL;
	rnd_hid_attribute_t *attr;
	rnd_hid_tree_t *tree;
	rnd_hid_row_t *r;
	char *cell[3], *cursor_path = NULL, tmp[256];
	rnd_hid_attr_val_t hv;
	osm_tags_t *tags = NULL;
	vtp0_t all = {0};
	long n;

	/* set title */
	if (ctx->commit_tags == NULL) {
		obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);
		if (obj != NULL) {
			sprintf(tmp, "%s #" OSM_PID, osm_obj_name[obj->hdr.type], obj->hdr.id);
			hv.str = tmp;
			tags = obj->hdr.calls->get_tags(obj);
		}
		else
			hv.str = "Object not available";
		rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wobjinfo, &hv);
	}
	else
		tags = ctx->commit_tags;

	/* reload the list */
	attr = &ctx->dlg[ctx->wlist];
	tree = attr->wdata;

	/* remember cursor */
	r = rnd_dad_tree_get_selected(attr);
	if (r != NULL)
		cursor_path = rnd_strdup(r->cell[0]);

	/* remove existing items */
	rnd_dad_tree_clear(tree);

	if (tags != NULL) {
		htsp_entry_t *e;

		for(e = htsp_first(tags); e != NULL; e = htsp_next(tags, e))
			vtp0_append(&all, e);

		qsort(all.array, all.used, sizeof(htsp_entry_t *), tag_cmp);

		cell[2] = NULL;
		for(n = 0; n < all.used; n++) {
			e = all.array[n];
			cell[0] = rnd_strdup(e->key);
			cell[1] = rnd_strdup(e->value);
			rnd_dad_tree_append(attr, NULL, cell);
		}
		vtp0_uninit(&all);

		/* restore cursor */
		if (cursor_path != NULL) {
			hv.str = cursor_path;
			rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wlist, &hv);
		}

	}

	free(cursor_path);
}

static void tag_select(rnd_hid_attribute_t *attrib, void *hid_ctx, rnd_hid_row_t *row)
{
	rnd_hid_tree_t *tree = attrib->wdata;
	tags_ctx_t *ctx = tree->user_ctx;
	const char *key = NULL, *val = NULL;
	rnd_hid_attr_val_t hv;

	if (row != NULL) {
		key = row->cell[0];
		val = row->cell[1];
	}

	TODO("set suggestions");

	hv.str = val;
	rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wval, &hv);
}

static void cb_set(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	tags_ctx_t *ctx = caller_data;
	osm_any_obj_t *obj = NULL;

	if (ctx->commit_tags == NULL)
		obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);

	if ((obj != NULL) || (ctx->commit_tags != NULL)) {
		rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wlist];
		rnd_hid_row_t *r = rnd_dad_tree_get_selected(attr);

		if (r == NULL) {
			rnd_message(RND_MSG_ERROR, "Select a tag first on the left side!\n");
			return;
		}

		if (ctx->commit_tags != NULL) {
			htsp_entry_t *e = htsp_getentry(ctx->commit_tags, r->cell[0]);
			if (e != NULL) {
				free(e->value);
				e->value = rnd_strdup(ctx->dlg[ctx->wval].val.str);
			}
		}
		else
			osm_obj_modify_tag(ctx->ly, obj, r->cell[0], ctx->dlg[ctx->wval].val.str, 1);

		tags_osm2dlg(ctx);
	}
}

static void cb_new(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	tags_ctx_t *ctx = caller_data;
	osm_any_obj_t *obj = NULL;

	if (ctx->commit_tags == NULL)
		obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);

	if ((obj != NULL) || (ctx->commit_tags != NULL)) {
		char *new_key, *new_val;
		new_key = rnd_hid_prompt_for(&osm.hidlib, "Key of the new tag", "", "Create tag");
		if (new_key == NULL)
			return;
		new_val = rnd_hid_prompt_for(&osm.hidlib, "Value of the new tag", "", "Create tag");
		if (new_val == NULL)
			return;

		if (ctx->commit_tags == NULL) {
			osm_tag_create(obj, new_key, new_val, 1);
			free(new_key);
		}
		else {
			osm_tag_set(ctx->commit_tags, new_key, new_val);
			/* ownership of both are passed to the hash */
		}

		tags_osm2dlg(ctx);
	}
}

static void cb_del(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr_btn)
{
	tags_ctx_t *ctx = caller_data;
	osm_any_obj_t *obj = NULL;


	if (ctx->commit_tags == NULL)
		obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);

	if ((obj != NULL) || (ctx->commit_tags != NULL)) {
		rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wlist];
		rnd_hid_row_t *r = rnd_dad_tree_get_selected(attr);

		if (r == NULL)
			return;

		if (ctx->commit_tags != NULL) {
			htsp_entry_t *e = htsp_popentry(ctx->commit_tags, r->cell[0]);
			free(e->key);
			free(e->value);
		}
		else
			osm_tag_remove(obj, r->cell[0], 1);

		tags_osm2dlg(ctx);
	}
}

static void cb_rename(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	tags_ctx_t *ctx = caller_data;
	osm_any_obj_t *obj = NULL;


/*	if (ctx->commit_tags != NULL) {
		htsp_entry_t *e;

		if (r == NULL)
			return;
		tags_osm2dlg(ctx);
	}
*/

	if (ctx->commit_tags == NULL)
		obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);

	if ((obj != NULL) || (ctx->commit_tags != NULL)) {
		rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wlist];
		rnd_hid_row_t *r = rnd_dad_tree_get_selected(attr);
		char *new_key;

		if (r == NULL)
			return;

		new_key = rnd_hid_prompt_for(&osm.hidlib, "New key for this tag", "", "Rename tag");
		if (strcmp(new_key, r->cell[0]) == 0)
			return;

		if (ctx->commit_tags == NULL) {
			uundo_freeze_serial(&ctx->ly->undo);

			osm_tag_create(obj, new_key, r->cell[1], 1);
			osm_tag_remove(obj, r->cell[0], 1);

			uundo_unfreeze_serial(&ctx->ly->undo);
			uundo_inc_serial(&ctx->ly->undo);
		}
		else {
			htsp_entry_t *e = htsp_popentry(ctx->commit_tags, r->cell[0]);
			htsp_set(ctx->commit_tags, new_key, e->value);
			free(e->key);
			/* ownerhsip of new_key and e->value are passed to the hash */
		}

		tags_osm2dlg(ctx);
	}

}


int osm_rnd_dlg_tags(osm_layer_t *ly, osm_objtype_t type, osm_id_t id, osm_tags_t *commit_tags)
{
	rnd_hid_dad_buttons_t clbtn_obj[] = {{"Close", 0}, {NULL, 0}};
	rnd_hid_dad_buttons_t clbtn_ci[] = {{"cancel", -1}, {"Commit!", 0}, {NULL, 0}};
	const char *lhdr[] = {"key", "value"};
	const char *shdr[] = {"value", "meaning"};
	const char *title;
	tags_ctx_t *ctx;
	int modal = (commit_tags != NULL), res = 0;

	if (commit_tags == NULL) {
		assert(id != 0);
		assert(type >= 0);
		assert(type < OSM_OBJ_max);
	}

	ctx = calloc(sizeof(tags_ctx_t), 1);
	ctx->ly = ly;
	ctx->obj_type = type;
	ctx->obj_id = id;
	ctx->commit_tags = commit_tags;

	RND_DAD_BEGIN_VBOX(ctx->dlg);
		RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

		if (commit_tags != NULL) {
			RND_DAD_TEXT(ctx->dlg, NULL);
				ctx->wcomment = RND_DAD_CURRENT(ctx->dlg);
		}

		RND_DAD_BEGIN_HPANE(ctx->dlg, "left-right");
			RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

			RND_DAD_BEGIN_VBOX(ctx->dlg); /* left */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

				RND_DAD_LABEL(ctx->dlg, "objinfo");
					ctx->wobjinfo = RND_DAD_CURRENT(ctx->dlg);

				RND_DAD_TREE(ctx->dlg, 2, 0, lhdr);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
					ctx->wlist = RND_DAD_CURRENT(ctx->dlg);
					RND_DAD_TREE_SET_CB(ctx->dlg, selected_cb, tag_select);
					RND_DAD_TREE_SET_CB(ctx->dlg, ctx, ctx);

			RND_DAD_END(ctx->dlg);

			RND_DAD_BEGIN_VBOX(ctx->dlg); /* right */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

				RND_DAD_LABEL(ctx->dlg, "Suggested values:");

				RND_DAD_TREE(ctx->dlg, 2, 0, shdr);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
					ctx->wsuggest = RND_DAD_CURRENT(ctx->dlg);

				RND_DAD_BEGIN_HBOX(ctx->dlg);
					RND_DAD_STRING(ctx->dlg);
						RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);
						ctx->wval = RND_DAD_CURRENT(ctx->dlg);
						RND_DAD_ENTER_CB(ctx->dlg, cb_set);
					RND_DAD_BUTTON(ctx->dlg, "set");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_set);

				RND_DAD_END(ctx->dlg);

				RND_DAD_BEGIN_HBOX(ctx->dlg); /* buttons */

					RND_DAD_BUTTON(ctx->dlg, "del");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_del);
					RND_DAD_BUTTON(ctx->dlg, "new");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_new);
					RND_DAD_BUTTON(ctx->dlg, "rename");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_rename);

					RND_DAD_BEGIN_VBOX(ctx->dlg);
						RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL); /* spring */
					RND_DAD_END(ctx->dlg);

					RND_DAD_BUTTON_CLOSES(ctx->dlg, (commit_tags != NULL) ? clbtn_ci : clbtn_obj);
				RND_DAD_END(ctx->dlg); /* buttons */
			RND_DAD_END(ctx->dlg); /* left */
		RND_DAD_END(ctx->dlg); /* pane */
	RND_DAD_END(ctx->dlg);


	if (commit_tags == NULL)
		title = "Object tags";
	else
		title = "Commit tags";

	RND_DAD_NEW("tags", ctx->dlg, title, ctx, modal, tags_close_cb);
	tags_osm2dlg(ctx);

	if (modal)
		res = RND_DAD_RUN(ctx->dlg);

	return res;
}

const char osm_acts_TagsDialog[] = "TagsDialog(object)\n";
const char osm_acth_TagsDialog[] = "Present the TagsDialog box of an object";
fgw_error_t osm_act_TagsDialog(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	osm_any_obj_t *obj;

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No active layer\n");
		return -1;
	}

	obj = osm_search_obj_xy(ly, osm->crosshair_x, osm->crosshair_y);
	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No object under cursor\n");
		return -1;
	}

	osm_rnd_dlg_tags(ly, obj->hdr.type, obj->hdr.id, NULL);
	RND_ACT_IRES(0);
	return 0;
}
