/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - Bing map tiles
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <ctype.h>
#include <genht/htsp.h>
#include <genht/hash.h>
#include <genvector/gds_char.h>
#include <librnd/core/compat_fs.h>
#include <librnd/core/safe_fs.h>
#include <librnd/core/paths.h>
#include <libxml/tree.h>
#include <libxml/parser.h>

#include "conf_core.h"
#include "src_plugins/lib_wgetosm/lib_wgetosm.h"
#include "src_plugins/lib_tilecache/lib_tilecache.h"

osm_tilecache_t bing_cache;

static char *bing_url_orig, *bing_url[3];
static vtp0_t bing_subdomains;
static int bing_subdomain_cnt;
static gds_t bing_tmp;
static int bing_api_ok;


TODO("these should be configurable, API key is coming from JOSM")
#define CACHE_DIR  "~/" DOT_OSM_RND "/cache/bing"
#define API_KEY    "key=Arzdiw4nlOJzRwOz__qailc8NiR31Tt51dN2D7cm57NrnceZnCpgOkmJhNpGoppU"
#define INIT_URL   "http://dev.virtualearth.net/REST/V1/Imagery/Metadata/Aerial?output=xml&include=ImageryProviders&" API_KEY

static xmlNode *search_xml_node(xmlNode *parent, const char *name)
{
	xmlNode *n;

	if (parent == NULL)
		return NULL;

	for(n = parent->children; n != NULL; n = n->next)
		if (xmlStrcmp(n->name, (xmlChar *)name) == 0)
			return n;

	return NULL;
}

static int bing_api_init(rnd_design_t *dsg)
{
	char *tmpfn;
	xmlDoc *doc;
	xmlNode *root, *rsc, *imeta, *nd, *n;
	int url_ok = 0, subdom_ok = 0;

	if (conf_core.editor.offline)
		return 0;

/* Doc on the API:
   https://learn.microsoft.com/en-us/bingmaps/rest-services/imagery/get-imagery-metadata
*/

#ifndef local_api_xml
	tmpfn = rnd_tempfile_name_new("bing_api.xml");
	rnd_message(RND_MSG_INFO, "Getting Bing API entry in %s...\n", tmpfn);
	osm_wget_disk(INIT_URL, tmpfn, 0, NULL);
#else
	tmpfn = local_api_xml;
#endif

	doc = xmlReadFile(tmpfn, NULL, 0);
	if (doc != NULL)
		root = xmlDocGetRootElement(doc);
	if (root != NULL) { /* root is Response */
		rsc = search_xml_node(root, "ResourceSets");
		rsc = search_xml_node(rsc, "ResourceSet");
		rsc = search_xml_node(rsc, "Resources");
		imeta = search_xml_node(rsc, "ImageryMetadata");
		nd = search_xml_node(imeta, "ImageUrl");
		if ((nd != NULL) && (nd->children != NULL)) {
			char *s = (char *)nd->children->content, *end, *sres, *sqk;
			while(isspace(*s)) s++;
			s = rnd_strdup(s);
			end = strpbrk(s, " \t\r\n");
			if (end != NULL)
				*end = '\0';

			bing_url_orig = s;
			sres = strstr(bing_url_orig, "{subdomain}");
			sqk = strstr(bing_url_orig, "{quadkey}");
			if ((sres != NULL) && (sqk != NULL)) {
				*sres = '\0';
				*sqk = '\0';
				bing_url[0] = bing_url_orig;
				bing_url[1] = sres+11;
				bing_url[2] = sqk+9;
				rnd_trace("Bing xml ImageUrl: '%s' '%s' '%s'\n", bing_url[0], bing_url[1], bing_url[2]);
				url_ok = 1;
			}
		}
		nd = search_xml_node(imeta, "ImageUrlSubdomains");
		if (nd != NULL) {
			for(n = nd->children; n != NULL; n = n->next) {
				if ((xmlStrcmp(n->name, (xmlChar *)"string") == 0) && (n->children != NULL)) {
					char *s = (char *)n->children->content, *end;
					while(isspace(*s)) s++;
					s = rnd_strdup(s);
					end = strpbrk(s, " \t\r\n");
					if (end != NULL)
						*end = '\0';
					vtp0_append(&bing_subdomains, s);
					rnd_trace("Bing xml subd: '%s'\n", s);
					subdom_ok = 1;
				}
			}
		}
	}
	xmlFreeDoc(doc);

#ifndef local_api_xml
	rnd_tempfile_unlink(tmpfn);
#endif

	if (!url_ok || !subdom_ok) {
		rnd_message(RND_MSG_ERROR, "Bing: failed to initialize the API; url template: %d, subdom: %d\n", url_ok, subdom_ok);
		return -1;
	}

	bing_api_ok = 1;
	bing_cache.offline = 0;
	return 0;
}

static const char *bing_build_url_for(rnd_design_t *dsg, osm_tilecache_t *ctx, const char *quadkey, void *tileaddr)
{
	const char *subd = bing_subdomains.array[(bing_subdomain_cnt++) % bing_subdomains.used];

	bing_tmp.used = 0;
	gds_append_str(&bing_tmp, bing_url[0]);
	gds_append_str(&bing_tmp, subd);
	gds_append_str(&bing_tmp, bing_url[1]);
	gds_append_str(&bing_tmp, quadkey);
	gds_append_str(&bing_tmp, bing_url[2]);

	return bing_tmp.array;
}

static void bing_snprint_disk_name(rnd_design_t *dsg, osm_tilecache_t *ctx, const char *quadkey, void *tileaddr, char *dst, size_t max_len)
{
	rnd_snprintf(ctx->cache_disk_path_end, max_len, "%s.jpg", quadkey);
}


static void bing_cache_init(rnd_design_t *dsg)
{
	if (bing_cache.inited == 0) {
		char *curr, *next;

		bing_cache.build_url_for = bing_build_url_for;
		bing_cache.snprint_disk_name = bing_snprint_disk_name;
		bing_cache.offline = 1; /* wait until the API is inited */

		osm_tilecache_init(&bing_cache);

		rnd_path_resolve(dsg, CACHE_DIR, &bing_cache.cache_disk_path, 256, 0);
		for(curr = bing_cache.cache_disk_path; curr != NULL; curr = next) {
			next = strchr(curr, '/');
			if (next != NULL)
				*next = '\0';
			rnd_mkdir(dsg, bing_cache.cache_disk_path, 0755);
			if (next != NULL) {
				*next = '/';
				next++;
			}
		}
		bing_cache.cache_disk_path_end = bing_cache.cache_disk_path + strlen(bing_cache.cache_disk_path);
		*bing_cache.cache_disk_path_end = '/';
		bing_cache.cache_disk_path_end++;

		bing_api_init(dsg);
	}
	bing_cache.inited++;
}

static rnd_pixmap_t *bing_cache_get(rnd_design_t *dsg, const char *quadkey, time_t now)
{
	return osm_tilecache_get(dsg, &bing_cache, quadkey, NULL, now);
}

