/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - OSM xml format
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <librnd/config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <genht/htsp.h>
#include <genht/hash.h>

#include "osm_typedefs.h"
#include "plug_io.h"
#include "data.h"
#include <librnd/core/error.h>
#include <librnd/core/safe_fs.h>
#include <librnd/core/compat_misc.h>
#include <librnd/core/misc_util.h>
#include "obj_any.h"

#include "io_osm_xml.h"

typedef struct {
	xmlDoc *doc;
	xmlNode *root;
	osm_design_t *osm;

	int is_gpx;
} read_ctx_t;

static int oxml_parse_tag(read_ctx_t *rctx, osm_any_obj_t *obj, xmlNode *nd)
{
	const char *k = (const char *)xmlGetProp(nd, (const xmlChar *)"k");
	const char *v = (const char *)xmlGetProp(nd, (const xmlChar *)"v");

	if ((k == NULL) || (v == NULL)) {
		rnd_message(RND_MSG_ERROR, "missing tag key or value\n");
		return -1;
	}

	osm_tag_set(obj->hdr.calls->get_tags(obj), k, v);

	return 0;
}

/* Returns 0 on error */
osm_id_t oxml_parse_obj_id(read_ctx_t *rctx, osm_layer_t *ly, xmlNode *nd)
{
	char *sid, *end;
	osm_id_t id;

	sid = (char *)xmlGetProp(nd, (const xmlChar *)"id");
	id = osm_str2id(sid, &end);
	if (*end != '\0') {
		rnd_message(RND_MSG_ERROR, "Missing or invalid object id: '%s'\n", sid);
		return 0;
	}

	if (id < ly->last_new_id)
		ly->last_new_id = id;

	return id;
}


static int oxml_parse_hdr(read_ctx_t *rctx, osm_any_obj_t *obj, xmlNode *nd, osm_id_t id)
{
	char *end;
	const char *s;
	int res = 0;

	obj->hdr.id = id;

	s = (char *)xmlGetProp(nd, (const xmlChar *)"uid");
	if (s != NULL) {
		obj->hdr.uid = osm_str2id(s, &end);
		if (*end != '\0') {
			rnd_message(RND_MSG_ERROR, "Invalid user id: '%s'\n", s);
			res = 1;
		}
	}
	else
		obj->hdr.uid = 0;

	TODO("load timestamp");

	s = (const char *)xmlGetProp(nd, (const xmlChar *)"visible");
	if ((s != NULL) && (rnd_strcasecmp(s, "false") == 0))
		obj->hdr.visible = 0;
	else
		obj->hdr.visible = 1;

	s = (char *)xmlGetProp(nd, (const xmlChar *)"version");
	if (s != NULL) {
		obj->hdr.version = osm_str2id(s, &end);
		if (*end != '\0') {
			rnd_message(RND_MSG_ERROR, "Invalid version: '%s'\n", s);
			res = 1;
		}
	}
	else
		obj->hdr.version = 0;


	s = (char *)xmlGetProp(nd, (const xmlChar *)"changeset");
	if (s != NULL) {
		obj->hdr.changeset = osm_str2id(s, &end);
		if (*end != '\0') {
			rnd_message(RND_MSG_ERROR, "Invalid changeset: '%s'\n", s);
			res = 1;
		}
	}
	else
		obj->hdr.changeset = 0;


	return res;
}

RND_INLINE int conv_lat_lon(const char *lats, const char *lons, double *lat_out, double *lon_out)
{
	double lat, lon;
	char *end;

	lon = strtod(lons, &end);
	if (*end != '\0') {
		rnd_message(RND_MSG_ERROR, "invalid numeric value in lon field\n");
		return -1;
	}
	if ((lon < -180) || (lon > 180)) {
		rnd_message(RND_MSG_ERROR, "invalid value in lon field (out of range)\n");
		return -1;
	}

	lat = strtod(lats, &end);
	if (*end != '\0') {
		rnd_message(RND_MSG_ERROR, "invalid numeric value in lon field\n");
		return -1;
	}
	if ((lat < -90) || (lat > 90)) {
		rnd_message(RND_MSG_ERROR, "invalid value in lat field (out of range)\n");
		return -1;
	}

	*lat_out = lat;
	*lon_out = lon;
	return 0;
}

static int oxml_parse_node(read_ctx_t *rctx, osm_layer_t *ly, xmlNode *nd)
{
	osm_node_t *node;
	const char *lons, *lats;
	double lon, lat;
	int res, mod = 0;
	xmlNode *n;
	osm_id_t id;

	id = oxml_parse_obj_id(rctx, ly, nd);
	if (id == 0)
		return -1;

	lons = (const char *)xmlGetProp(nd, (const xmlChar *)"lon");
	lats = (const char *)xmlGetProp(nd, (const xmlChar *)"lat");
	if (conv_lat_lon(lats, lons, &lat, &lon) != 0)
		return -1;

	node = htIp_get(&ly->id2obj[OSM_OBJ_NODE], id);
	if (node != NULL) {
		osm_obj_pre_edit_orig((osm_any_obj_t *)node, 1);
		mod = 1;
	}
	else
		node = osm_node_new();

	node->orig.lon = lon;
	node->orig.lat = lat;

	res = oxml_parse_hdr(rctx, (osm_any_obj_t *)node, nd, id);

	for(n = nd->children; n != NULL; n = n->next)
		if (xmlStrcmp(n->name, (xmlChar *)"tag") == 0)
			res |= oxml_parse_tag(rctx, (osm_any_obj_t *)node, n);


	if (mod)
		osm_obj_post_edit_orig((osm_any_obj_t *)node, 1);
	else
		osm_obj_add_to_layer(ly, (osm_any_obj_t *)node);

	return res;
}

static int oxml_parse_way(read_ctx_t *rctx, osm_layer_t *ly, xmlNode *nd)
{
	osm_way_t *way;
	char *end;
	int res, mod = 0;
	xmlNode *n;
	osm_id_t id;

	id = oxml_parse_obj_id(rctx, ly, nd);
	if (id == 0)
		return -1;

	way = htIp_get(&ly->id2obj[OSM_OBJ_WAY], id);
	if (way != NULL) {
		osm_obj_pre_edit_orig((osm_any_obj_t *)way, 1);
		mod = 1;
	}
	else
		way = osm_way_new();

	way->orig.nd.used = 0;
	for(n = nd->children; n != NULL; n = n->next) {
		if (xmlStrcmp(n->name, (xmlChar *)"nd") == 0) {
			osm_id_t ref = osm_str2id((char *)xmlGetProp(n, (const xmlChar *)"ref"), &end);
			if (*end == '\0')
				vtI0_append(&way->orig.nd, ref);
			else
				res = -1;
		}
		else if (xmlStrcmp(n->name, (xmlChar *)"tag") == 0)
			res |= oxml_parse_tag(rctx, (osm_any_obj_t *)way, n);
	}

	res = oxml_parse_hdr(rctx, (osm_any_obj_t *)way, nd, id);

	if (mod)
		osm_obj_post_edit_orig((osm_any_obj_t *)way, 1);
	else
		osm_obj_add_to_layer(ly, (osm_any_obj_t *)way);

	return res;
}

static int oxml_parse_relation(read_ctx_t *rctx, osm_layer_t *ly, xmlNode *nd)
{
	osm_relation_t *relation;
	char *end;
	long i;
	int res, mod = 0;
	xmlNode *n;
	osm_id_t id;

	id = oxml_parse_obj_id(rctx, ly, nd);
	if (id == 0)
		return -1;

	relation = htIp_get(&ly->id2obj[OSM_OBJ_RELATION], id);
	if (relation != NULL) {
		osm_obj_pre_edit_orig((osm_any_obj_t *)relation, 1);
		mod = 1;
	}
	else
		relation = osm_relation_new();

	for(i = 0; i < relation->orig.members.used; i++)
		osm_relmember_free_fields(&relation->orig.members.array[i]);
	relation->orig.members.used = 0;

	for(n = nd->children; n != NULL; n = n->next) {
		if (xmlStrcmp(n->name, (xmlChar *)"member") == 0) {
			osm_id_t ref = osm_str2id((char *)xmlGetProp(n, (const xmlChar *)"ref"), &end);
			xmlChar *stype = xmlGetProp(n, (const xmlChar *)"type");
			const char *srole = (const char *)xmlGetProp(n, (const xmlChar *)"role");
			osm_objtype_t type = OSM_OBJ_invalid;

			if (xmlStrcmp(stype, (xmlChar *)"node") == 0) type = OSM_OBJ_NODE;
			else if (xmlStrcmp(stype, (xmlChar *)"way") == 0) type = OSM_OBJ_WAY;
			else if (xmlStrcmp(stype, (xmlChar *)"relation") == 0) type = OSM_OBJ_RELATION;

			if ((*end == '\0') && (type != OSM_OBJ_invalid)) {
				osm_relmember_t *rm = vtr0_alloc_append(&relation->orig.members, 1);
				rm->type = type;
				rm->ref = ref;
				rm->role = srole == NULL ? NULL : rnd_strdup(srole);
			}
			else
				res = -1;
		}
		else if (xmlStrcmp(n->name, (xmlChar *)"tag") == 0)
			res |= oxml_parse_tag(rctx, (osm_any_obj_t *)relation, n);
	}

	res = oxml_parse_hdr(rctx, (osm_any_obj_t *)relation, nd, id);

	if (mod)
		osm_obj_post_edit_orig((osm_any_obj_t *)relation, 1);
	else
		osm_obj_add_to_layer(ly, (osm_any_obj_t *)relation);

	return res;
}


static int oxml_parse_layer_osm(read_ctx_t *rctx, osm_layer_t *ly)
{
	xmlNode *n;
	int res = 0;

	for(n = rctx->root->children; n != NULL; n = n->next) {
		if (xmlStrcmp(n->name, (xmlChar *)"node") == 0)
			res |= oxml_parse_node(rctx, ly, n);
		else if (xmlStrcmp(n->name, (xmlChar *)"way") == 0)
			res |= oxml_parse_way(rctx, ly, n);
		else if (xmlStrcmp(n->name, (xmlChar *)"relation") == 0)
			res |= oxml_parse_relation(rctx, ly, n);
	}

	ly->editable = 1;

	return res;
}

/* Returns whether a <trkpt> has a child called <time> */
RND_INLINE int trkpt_has_time(xmlNode *trkpt)
{
	xmlNode *n;

	for(n = trkpt->children; n != NULL; n = n->next)
		if (xmlStrcmp(n->name, (xmlChar *)"time") == 0)
			return 1;

	return 0;
}

static int oxml_parse_trkseg(read_ctx_t *rctx, osm_layer_t *ly, xmlNode *nd, osm_way_t **way_out)
{
	osm_way_t *way;
	char *end;
	xmlNode *n;

	way = osm_way_new();
	way->hdr.id = - (--ly->last_new_id); /* use positive ids as these are not new objects */
	way->hdr.version = 1<<30; /* so this can not be commited ever */

	way->orig.nd.used = 0;
	for(n = nd->children; n != NULL; n = n->next) {
		if (n->type == XML_TEXT_NODE) continue;
		if (xmlStrcmp(n->name, (xmlChar *)"trkpt") == 0) {
			const char *lats, *lons;
			double lat, lon;
			osm_node_t *node;

			lons = (const char *)xmlGetProp(n, (const xmlChar *)"lon");
			lats = (const char *)xmlGetProp(n, (const xmlChar *)"lat");
			if (conv_lat_lon(lats, lons, &lat, &lon) != 0)
				return -1;

			node = osm_node_new();
			node->hdr.id = - (--ly->last_new_id); /* use positive ids as these are not new objects */
			node->hdr.version = 1<<30; /* so this can not be commited ever */
			node->orig.lat = lat;
			node->orig.lon = lon;

			osm_obj_add_to_layer(ly, (osm_any_obj_t *)node);

			/* trkpt without time is anonymized, sorted by lat/lon, can't be used in a way */
			if (trkpt_has_time(n))
				vtI0_append(&way->orig.nd, node->hdr.id);
		}
	}

	if (way->orig.nd.used > 1) {
		osm_obj_add_to_layer(ly, (osm_any_obj_t *)way);
		*way_out = way;
	}
	else {
		osm_obj_free((osm_any_obj_t *)way);
		*way_out = NULL;
	}

	return 0;
}

static int oxml_parse_layer_gpx(read_ctx_t *rctx, osm_layer_t *ly)
{
	xmlNode *n, *s;
	int res = 0;
	vtp0_t ways = {0};

	ly->editable = 0;
	ly->force_style_name = "_gpx";

	for(n = rctx->root->children; n != NULL; n = n->next) {
		if (n->type == XML_TEXT_NODE) continue;

		if (xmlStrcmp(n->name, (xmlChar *)"trk") == 0) {

			/* read all ways */
			ways.used = 0;
			for(s = n->children; s != NULL; s = s->next) {
				if (s->type == XML_TEXT_NODE) continue;
				if (xmlStrcmp(s->name, (xmlChar *)"trkseg") == 0) {
					osm_way_t *way;
					res |= oxml_parse_trkseg(rctx, ly, s, &way);
					if (way != NULL)
						vtp0_append(&ways, way);
				}
			}

			if (ways.used > 0) {
				/* read metadata and tag all ways within trk */
				for(s = n->children; s != NULL; s = s->next) {
					if (s->type == XML_TEXT_NODE) continue;
					if (xmlStrcmp(s->name, (xmlChar *)"trkseg") != 0) {
						long i;
						const char *val = (const char *)xmlNodeGetContent(s->children);
						if (val == NULL) val = "";
						
						for(i = 0; i < ways.used; i++) {
							osm_way_t *way = ways.array[i];
							osm_tag_set(&way->orig.tags, (const char *)s->name, val);
						}
					}
				}
			}
		}
	}

	return res;
}


static int oxml_parse_layer(read_ctx_t *rctx, osm_layer_t *ly)
{
	return rctx->is_gpx ? oxml_parse_layer_gpx(rctx, ly) : oxml_parse_layer_osm(rctx, ly);
}


static int load_xml(read_ctx_t *rctx, const char *fn)
{
	FILE *f;
	char *efn = NULL;

	f = rnd_fopen_fn(NULL, fn, "r", &efn);
	if (f == NULL) {
		rnd_message(RND_MSG_ERROR, "can't open '%s'\n", fn);
		return -1;
	}
	fclose(f);

	rctx->doc = xmlReadFile(efn, NULL, 0);
	if (rctx->doc == NULL) {
		rnd_message(RND_MSG_ERROR, "xml parsing error on file %s (%s)\n", fn, efn);
		free(efn);
		return -1;
	}
	free(efn);

	rctx->root = xmlDocGetRootElement(rctx->doc);

	if (xmlStrcmp(rctx->root->name, (xmlChar *)"osm") == 0) {
		rctx->is_gpx = 0;
		return 0;
	}

	if (xmlStrcmp(rctx->root->name, (xmlChar *)"gpx") == 0) {
		rctx->is_gpx = 1;
		return 0;
	}

	rnd_message(RND_MSG_ERROR, "xml error: root is neither <osm> nor <gpx>\n");
	xmlFreeDoc(rctx->doc);
	return -1;
}

static void free_xml(read_ctx_t *rctx)
{
	xmlFreeDoc(rctx->doc);
	rctx->doc = NULL;
	rctx->root = NULL;
}

RND_INLINE int osm_oxml_load_(osm_design_t *osm, const char *fn, osm_layer_t **ly)
{
	int res = 0;
	read_ctx_t rctx = {0};

	if (load_xml(&rctx, fn) != 0)
		return -1;

	if (ly != NULL) {
		if (*ly == NULL)
			*ly = osm_layer_by_name(osm, fn, 1);
		res = oxml_parse_layer(&rctx, *ly);
	}
	else {
		osm_layer_t *lytmp = osm_layer_by_name(osm, fn, 1);
		res = oxml_parse_layer(&rctx, lytmp);
	}

	free_xml(&rctx);
	return res;
}

static int osm_oxml_load(osm_design_t *osm, const char *fn, FILE *f)
{
	return osm_oxml_load_(osm, fn, NULL);
}


int osm_oxml_load_into(osm_design_t *osm, const char *fn, osm_layer_t **ly)
{
	return osm_oxml_load_(osm, fn, ly);
}


/*** write ***/


#define Q     "\""
#define PID   Q OSM_PID Q

void osm_xml_fprintf_qs(FILE *f, const char *s)
{
	for(; *s != '\0'; s++) {
		if (*s == '&')
			fprintf(f, "&amp;");
		else if (*s == '"')
			fprintf(f, "&quot;");
		else
			fputc(*s, f);
	}
}

RND_INLINE void osm_oxml_fprint_hdr(FILE *f, const char *ind, const char *name, const osm_any_obj_t *obj, const osm_id_t *changeset_id)
{
	const char *viss = obj->hdr.visible ? Q "true" Q : Q "false" Q;
	osm_id_t chs = changeset_id == NULL ? obj->hdr.changeset : *changeset_id;

	fprintf(f, "%s<%s id="PID" visible=%s version=\"%ld\" changeset="PID, ind, name, obj->hdr.id, viss, obj->hdr.version, chs);
TODO("save timestamp");
}

RND_INLINE void osm_oxml_fprint_tags(FILE *f, const char *ind, const osm_tags_t *tags)
{
	htsp_entry_t *e;
	for(e = htsp_first(tags); e != NULL; e = htsp_next(tags, e)) {
		fprintf(f, "%s <tag k=\"", ind);
		osm_xml_fprintf_qs(f, e->key);
		fprintf(f, "\" v=\"");
		osm_xml_fprintf_qs(f, (char *)e->value);
		fprintf(f, "\"/>\n");
		
	}
}


RND_INLINE void osm_oxml_fprint_node(FILE *f, const char *ind, const osm_any_obj_t *obj, const osm_id_t *changeset_id)
{
	osm_oxml_fprint_hdr(f, ind, "node", obj, changeset_id);
	fprintf(f, " lat=\"%.7f\" lon=\"%.7f\">\n", obj->node.curr->lat, obj->node.curr->lon);

	osm_oxml_fprint_tags(f, ind, &obj->node.curr->tags);
	fprintf(f, "%s</node>\n", ind);
}

RND_INLINE void osm_oxml_fprint_way(FILE *f, const char *ind, const osm_any_obj_t *obj, const osm_id_t *changeset_id)
{
	long n;

	osm_oxml_fprint_hdr(f, ind, "way", obj, changeset_id);
	fprintf(f, ">\n");

	for(n = 0; n < obj->way.curr->nd.used; n++)
		fprintf(f, "%s <nd ref="PID"/>\n", ind, obj->way.curr->nd.array[n]);

	osm_oxml_fprint_tags(f, ind, &obj->way.curr->tags);
	fprintf(f, "%s</way>\n", ind);
}

RND_INLINE void osm_oxml_fprint_relation(FILE *f, const char *ind, const osm_any_obj_t *obj, const osm_id_t *changeset_id)
{
	long n;

	osm_oxml_fprint_hdr(f, ind, "relation", obj, changeset_id);
	fprintf(f, ">\n");

	for(n = 0; n < obj->relation.curr->members.used; n++) {
		osm_relmember_t *rm = &(obj->relation.curr->members.array[n]);
		const char *role = rm->role == NULL ? "" : rm->role;

		assert(rm->type >= 0);
		assert(rm->type < OSM_OBJ_max);
		fprintf(f, "%s <member type=\"%s\" ref="PID" role=\"", ind, osm_obj_name[rm->type], rm->ref);
		osm_xml_fprintf_qs(f, role);
		fprintf(f, "\"/>\n");
	}

	osm_oxml_fprint_tags(f, ind, &obj->relation.curr->tags);
	fprintf(f, "%s</relation>\n", ind);
}

void osm_oxml_fprint_obj(FILE *f, const char *ind, const osm_any_obj_t *obj, const osm_id_t *changeset_id)
{
	switch(obj->hdr.type) {
		case OSM_OBJ_NODE:     osm_oxml_fprint_node(f, ind, obj, changeset_id); return;
		case OSM_OBJ_WAY:      osm_oxml_fprint_way(f, ind, obj, changeset_id); return;
		case OSM_OBJ_RELATION: osm_oxml_fprint_relation(f, ind, obj, changeset_id); return;

		case OSM_OBJ_invalid: case OSM_OBJ_max:
			abort();
	}

	/* can't get here */
	abort();
}


typedef struct {
	osm_design_t *osm;
	FILE *f;
} write_ctx_t;

static int osm_oxml_save_layer(osm_design_t *osm, osm_layer_t *ly, const char *fn)
{
	write_ctx_t wctx = {0};
	int t;
	htIp_entry_t *e;

	wctx.osm = osm;
	wctx.f = rnd_fopen(&osm->hidlib, fn, "w");
	if (wctx.f == NULL) {
		rnd_message(RND_MSG_ERROR, "Failed to open %s for write\n", fn);
		return -1;
	}

	fprintf(wctx.f, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	fprintf(wctx.f, "<osm version=\"0.6\" generator=\"osm-rnd " OSM_VERS " osm_oxml_save_layer()\">\n");

	for(t = 0; t < OSM_OBJ_max; t++)
		for(e = htIp_first(&ly->id2obj[t]); e != NULL; e = htIp_next(&ly->id2obj[t], e))
			osm_oxml_fprint_obj(wctx.f, " ", e->value, NULL);

	fprintf(wctx.f, "</osm>\n");
	fclose(wctx.f);
	return 0;
}


static int osm_oxml_save_design(osm_design_t *osm, const char *fn)
{
	TODO("implement me");
	return -1;
}

static int osm_oxml_test_load(osm_design_t *osm, const char *fn, FILE *f)
{
	char *line, line_[1024];
	int n = 0;

	while((line = fgets(line_, sizeof(line_), f)) != NULL) {
		char *so, *sg;
		
		so = strstr(line, "<osm");
		if ((so != NULL) && (isspace(so[4])))
			return 1;

		sg = strstr(line, "<gpx");
		if ((sg != NULL) && (isspace(sg[4])))
			return 1;

		if (n > 32)
			return 0;
	}
	return 0;
}

static int osm_oxml_save_prio(const char *fn, const char *fmt, osm_io_type_t type)
{
	if (((type & OSM_IOTYP_DESIGN) == 0) && ((type & OSM_IOTYP_LAYER) == 0))
		return -1;

	if ((rnd_strcasecmp(fmt, "xml") != 0) && (rnd_strcasecmp(fmt, "osm_xml") != 0))
		return -1;

	return 100;
}

static osm_io_t io_osm_xml = {
	"osm xml", 100,
	osm_oxml_test_load,
	osm_oxml_load,
	osm_oxml_save_prio,
	osm_oxml_save_design,
	osm_oxml_save_layer,
	"osm_xml",
	"xml"
};


int pplg_check_ver_io_osm_xml(int ver_needed) { return 0; }

void pplg_uninit_io_osm_xml(void)
{
	osm_io_unreg(&io_osm_xml);
}

int pplg_init_io_osm_xml(void)
{
	osm_io_reg(&io_osm_xml);
	return 0;
}
