/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - generic tile cache
 *  Copyright (C) 2023,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include "conf_core.h"
#include <librnd/core/plugins.h>
#include <librnd/core/safe_fs.h>
#include <librnd/core/compat_misc.h>
#include <librnd/core/error.h>
#include <genht/hash.h>

#include "src_plugins/lib_wgetosm/lib_wgetosm.h"

#include "data.h"
#include "lib_tilecache.h"

/* This is generic and could be moved to a lib */
RND_INLINE void pixmap_mirror_y(rnd_pixmap_t *pm)
{
	unsigned char *r1, *r2;
	int x, y;

	assert(pm->sx == 256);
	assert(pm->sy == 256);
	r1 = pm->p;
	r2 = pm->p + 255*256*3;

	for(y = 0; y < 128; y++) {
		unsigned char *p1 = r1, *p2 = r2;
		for(x = 0; x < 256*3; x++, p1++, p2++) {
			unsigned char tmp = *p1;
			*p1 = *p2;
			*p2 = tmp;
		}
		r1 += 256*3;
		r2 -= 256*3;
	}
}

static osm_tile_t *osm_cache_load_from_disk(rnd_design_t *dsg, osm_tilecache_t *ctx, const char *tilekey, void *tileaddr)
{
	osm_tile_t *res = NULL;
	rnd_pixmap_t *pm;

	ctx->snprint_disk_name(dsg, ctx, tilekey, tileaddr, ctx->cache_disk_path_end, 256);

	pm = rnd_pixmap_load(dsg, ctx->cache_disk_path);
	rnd_trace("cache load: '%s' -> %p\n", ctx->cache_disk_path, pm);


	if (pm != NULL) {
		pixmap_mirror_y(pm);
		res = malloc(sizeof(osm_tile_t));
		res->pm = pm;
		res->tilekey = rnd_strdup(tilekey);
		htsp_set(&ctx->cache, res->tilekey, res);
		TODO("set res->mtime");
	}

	*ctx->cache_disk_path_end = '\0';
	return res;
}

static osm_tile_t *osm_cache_load_from_net(rnd_design_t *dsg, osm_tilecache_t *ctx, const char *tilekey, void *tileaddr)
{
	osm_tile_t *res = NULL;
	const char *url = ctx->build_url_for(dsg, ctx, tilekey, tileaddr);

	ctx->snprint_disk_name(dsg, ctx, tilekey, tileaddr, ctx->cache_disk_path_end, 256);

	rnd_trace("@@ load from net: '%s': '%s' -> '%s'\n", tilekey, url, ctx->cache_disk_path);
	if (osm_wget_disk(url, ctx->cache_disk_path, 1, NULL) == 0)
		res = osm_cache_load_from_disk(dsg, ctx, tilekey, tileaddr);

	*ctx->cache_disk_path_end = '\0';

	return res;
}

rnd_pixmap_t *osm_tilecache_get(rnd_design_t *dsg, osm_tilecache_t *ctx, const char *tilekey, void *tileaddr, time_t now)
{
	osm_tile_t *tile = htsp_get(&ctx->cache, tilekey);
	if (tile == NULL) {
		tile = osm_cache_load_from_disk(dsg, ctx, tilekey, tileaddr);
		if ((tile == NULL) && !ctx->offline && !conf_core.editor.offline)  {
			TODO("when async wget is available: schedule download if not already scheduled by creating a tile with ->pm == NULL");
			tile = osm_cache_load_from_net(dsg, ctx, tilekey, tileaddr);
		}
	}
	if (tile == NULL)
		return NULL;
	tile->last_drawn = now;
	return tile->pm;
}

void osm_tilecache_init(osm_tilecache_t *ctx)
{
	htsp_init(&ctx->cache, strhash, strkeyeq);
}


int pplg_check_ver_lib_tilecache(int ver_needed) { return 0; }

void pplg_uninit_lib_tilecache(void)
{
}

int pplg_init_lib_tilecache(void)
{
	RND_API_CHK_VER;

	return 0;
}
