/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - Optimizations
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <math.h>
#include <genlist/gendlist.h>
#include <librnd/config.h>
#include <librnd/core/actions.h>
#include <librnd/core/rotate.h>
#include <librnd/hid/hid.h>

#include "data.h"
#include "search.h"
#include "obj_node.h"
#include "obj_way.h"
#include "projection.h"
#include "geo.h"
#include "draw.h"

static const char optimize_cookie[] = "optimize";

#define MARK referer

/*** Implementation ***/

#include "ortho.c"
#include "lineup.c"

/*** Entry points ***/

const char osm_acts_Orthogonalize[] = "Orthogonalize()";
const char osm_acth_Orthogonalize[] = "Tune way angles of selected ways to have right angle corners. If 2 nodes are also selected, use the vector between them to set the preferred orientation\n";
fgw_error_t osm_act_Orthogonalize(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	vtp0_t way_list = {0};
	osm_node_t *dir[2] = {0};

	if ((ly == NULL) || !ly->editable) {
		rnd_message(RND_MSG_ERROR, "Select an editable layer first!\n");
		return -1;
	}

	osm_opt_ortho_search(ly, &way_list, dir);

# if 0
	{ /* debug print results */
		long n;
		rnd_trace("opt search res:\n");
		for(n = 0; n < way_list.used; n++) {
			osm_way_t *way = way_list.array[n];
			rnd_trace(" w " OSM_PID "\n", way->hdr.id);
		}
		if (dir[0] != NULL) {
			rnd_trace(" n " OSM_PID "\n", dir[0]->hdr.id);
			rnd_trace(" n " OSM_PID "\n", dir[1]->hdr.id);
		}
	}
#endif

	osm_draw_inhibit_inc();
	osm_opt_ortho(ly, &way_list, dir);
	osm_draw_inhibit_dec();

	vtp0_uninit(&way_list);
	return 0;
}

const char osm_acts_Lineup[] = "Lineup()";
const char osm_acth_Lineup[] = "Move a set of points to line up between the endpoints (which are not moved).\n";
fgw_error_t osm_act_Lineup(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	vtp0_t node_list = {0};

	if ((ly == NULL) || !ly->editable) {
		rnd_message(RND_MSG_ERROR, "Select an editable layer first!\n");
		return -1;
	}

	osm_opt_lineup_search(ly, &node_list);

	osm_draw_inhibit_inc();
	osm_opt_lineup(ly, &node_list);
	osm_draw_inhibit_dec();

	vtp0_uninit(&node_list);
	return 0;
}


static rnd_action_t optimize_action_list[] = {
	{"orthogonalize", osm_act_Orthogonalize, osm_acth_Orthogonalize, osm_acts_Orthogonalize},
	{"lineup", osm_act_Lineup, osm_acth_Lineup, osm_acts_Lineup}
};

int pplg_check_ver_optimize(int ver_needed) { return 0; }

void pplg_uninit_optimize(void)
{
	rnd_remove_actions_by_cookie(optimize_cookie);
}

int pplg_init_optimize(void)
{
	RND_REGISTER_ACTIONS(optimize_action_list, optimize_cookie);

	return 0;
}
